# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random
import string

from arrow import utcnow
from kombu import Connection

from swh.scheduler.updater.ghtorrent import RabbitMQConn, events


class FakeRandomOriginGenerator:
    def _random_string(self, length):
        """Build a fake string of length length.

        """
        return ''.join([
            random.choice(string.ascii_letters + string.digits)
            for n in range(length)])

    def generate(self, user_range=range(5, 10), repo_range=range(10, 15)):
        """Build a fake url

        """
        length_username = random.choice(user_range)
        user = self._random_string(length_username)
        length_repo = random.choice(repo_range)
        repo = self._random_string(length_repo)
        return '%s/%s' % (user, repo)


class FakeGHTorrentPublisher(RabbitMQConn):
    """Fake GHTorrent that randomly publishes fake events.  Those events
    are published in similar manner as described by ghtorrent's
    documentation [2].

    context: stuck with raw ghtorrent so far [1]

    [1] https://github.com/ghtorrent/ghtorrent.org/issues/397#issuecomment-387052462  # noqa
    [2] http://ghtorrent.org/streaming.html

    """

    ADDITIONAL_CONFIG = {
        'nb_messages': ('int', 100)
    }

    def __init__(self, **config):
        super().__init__(**config)
        self.fake_origin_generator = FakeRandomOriginGenerator()
        self.nb_messages = self.config['nb_messages']

    def _random_event(self):
        """Create a fake and random event

        """
        event_type = random.choice(['evt', 'ent'])
        sub_event = random.choice(events[event_type])
        return {
            'type': sub_event,
            'repo': {
                'name': self.fake_origin_generator.generate(),
            },
            'created_at': utcnow().isoformat()

        }

    def publish(self, nb_messages=None):
        if not nb_messages:
            nb_messages = self.nb_messages

        with Connection(self.config['conn']['url']) as conn:
            with conn.Producer(serializer='json') as producer:
                for n in range(nb_messages):
                    event = self._random_event()
                    producer.publish(event,
                                     exchange=self.exchange,
                                     routing_key=self.routing_key,
                                     declare=[self.queue])



# Copyright (C) 2016-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import arrow
import click
import csv
import itertools
import json
import locale
import logging

from swh.core import utils
from . import compute_nb_tasks_from
from .backend_es import SWHElasticSearchClient


locale.setlocale(locale.LC_ALL, '')
ARROW_LOCALE = locale.getlocale(locale.LC_TIME)[0]


class DateTimeType(click.ParamType):
    name = 'time and date'

    def convert(self, value, param, ctx):
        if not isinstance(value, arrow.Arrow):
            value = arrow.get(value)

        return value


DATETIME = DateTimeType()
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


def pretty_print_list(list, indent):
    """Pretty-print a list"""
    return ''.join('%s%s\n' % (' ' * indent, item) for item in list)


def pretty_print_dict(dict, indent):
    """Pretty-print a list"""
    return ''.join('%s%s: %s\n' %
                   (' ' * indent, click.style(key, bold=True), value)
                   for key, value in dict.items())


def pretty_print_task(task):
    """Pretty-print a task"""
    next_run = arrow.get(task['next_run'])
    lines = [
        '%s %s\n' % (click.style('Task', bold=True), task['id']),
        click.style('  Next run: ', bold=True),
        "%s (%s)" % (next_run.humanize(locale=ARROW_LOCALE),
                     next_run.format()),
        '\n',
        click.style('  Interval: ', bold=True),
        str(task['current_interval']), '\n',
        click.style('  Type: ', bold=True), task['type'], '\n',
        click.style('  Policy: ', bold=True), task['policy'], '\n',
        click.style('  Args:\n', bold=True),
        pretty_print_list(task['arguments']['args'], indent=4),
        click.style('  Keyword args:\n', bold=True),
        pretty_print_dict(task['arguments']['kwargs'], indent=4),
    ]

    return ''.join(lines)


def list_task_types(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo("Known task types:")
    for tasktype in ctx.obj.get_task_types():
        click.echo('{type}:\n  {description}'.format(**tasktype))
    ctx.exit()


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option('--cls', '-c', default='local',
              help="Scheduler's class, default to 'local'")
@click.option('--database', '-d',
              help='Scheduling database DSN')
@click.option('--url', '-u',
              help="(Optional) Scheduler's url access")
@click.pass_context
def cli(ctx, cls, database, url):
    """Software Heritage Scheduler CLI interface

    Default to use the the local scheduler instance (plugged to the
    main scheduler db).

    """
    scheduler = None
    override_config = {}
    from . import get_scheduler
    if cls == 'local':
        if database:
            override_config = {'scheduling_db': database}
        scheduler = get_scheduler(cls, args=override_config)
    elif cls == 'remote':
        if url:
            override_config = {'url': url}
        scheduler = get_scheduler(cls, args=override_config)

    if not scheduler:
        raise ValueError('Scheduler class (local/remote) must be instantiated')

    ctx.obj = scheduler


@cli.group('task')
@click.option('--list-types', '-l', is_flag=True, default=False, is_eager=True,
              expose_value=False, callback=list_task_types)
@click.pass_context
def task(ctx):
    """Manipulate tasks."""
    pass


@task.command('schedule')
@click.option('--columns', '-c', multiple=True,
              default=['type', 'args', 'kwargs', 'next_run'],
              type=click.Choice([
                  'type', 'args', 'kwargs', 'policy', 'next_run']),
              help='columns present in the CSV file')
@click.option('--delimiter', '-d', default=',')
@click.argument('file', type=click.File(encoding='utf-8'))
@click.pass_context
def schedule_tasks(ctx, columns, delimiter, file):
    """Schedule tasks from a CSV input file.

    The following columns are expected, and can be set through the -c option:

     - type: the type of the task to be scheduled (mandatory)

     - args: the arguments passed to the task (JSON list, defaults to an empty
       list)

     - kwargs: the keyword arguments passed to the task (JSON object, defaults
       to an empty dict)

     - next_run: the date at which the task should run (datetime, defaults to
       now)

    The CSV can be read either from a named file, or from stdin (use - as
    filename).

    Use sample:

    cat scheduling-task.txt | \
        python3 -m swh.scheduler.cli \
            --database 'service=swh-scheduler-dev' \
            task schedule \
                --columns type --columns kwargs --columns policy \
                --delimiter ';' -

    """
    tasks = []
    now = arrow.utcnow()

    reader = csv.reader(file, delimiter=delimiter)
    for line in reader:
        task = dict(zip(columns, line))
        args = json.loads(task.pop('args', '[]'))
        kwargs = json.loads(task.pop('kwargs', '{}'))
        task['arguments'] = {
            'args': args,
            'kwargs': kwargs,
        }
        task['next_run'] = DATETIME.convert(task.get('next_run', now),
                                            None, None)
        tasks.append(task)

    created = ctx.obj.create_tasks(tasks)

    output = [
        'Created %d tasks\n' % len(created),
    ]
    for task in created:
        output.append(pretty_print_task(task))

    click.echo_via_pager('\n'.join(output))


@task.command('add')
@click.argument('type', nargs=1, required=True)
@click.argument('options', nargs=-1)
@click.option('--policy', '-p', default='recurring',
              type=click.Choice(['recurring', 'oneshot']))
@click.option('--next-run', '-n', default=None)
@click.pass_context
def schedule_task(ctx, type, options, policy, next_run):
    """Schedule one task from arguments.

    Use sample:

    swh-scheduler --database 'service=swh-scheduler' \
        task add swh-lister-pypi

    swh-scheduler --database 'service=swh-scheduler' \
        task add swh-lister-debian --policy=oneshot distribution=stretch

    """
    now = arrow.utcnow()

    args = [x for x in options if '=' not in x]
    kw = dict(x.split('=', 1) for x in options if '=' in x)
    task = {'type': type,
            'policy': policy,
            'arguments': {
                'args': args,
                'kwargs': kw,
                },
            'next_run': DATETIME.convert(next_run or now,
                                         None, None),
            }
    created = ctx.obj.create_tasks([task])

    output = [
        'Created %d tasks\n' % len(created),
    ]
    for task in created:
        output.append(pretty_print_task(task))

    click.echo('\n'.join(output))


@task.command('list-pending')
@click.argument('task-types', required=True, nargs=-1)
@click.option('--limit', '-l', required=False, type=click.INT,
              help='The maximum number of tasks to fetch')
@click.option('--before', '-b', required=False, type=DATETIME,
              help='List all jobs supposed to run before the given date')
@click.pass_context
def list_pending_tasks(ctx, task_types, limit, before):
    """List the tasks that are going to be run.

    You can override the number of tasks to fetch

    """
    num_tasks, num_tasks_priority = compute_nb_tasks_from(limit)

    output = []
    for task_type in task_types:
        pending = ctx.obj.peek_ready_tasks(
            task_type, timestamp=before,
            num_tasks=num_tasks, num_tasks_priority=num_tasks_priority)
        output.append('Found %d %s tasks\n' % (
            len(pending), task_type))

        for task in pending:
            output.append(pretty_print_task(task))

    click.echo('\n'.join(output))


@task.command('archive')
@click.option('--before', '-b', default=None,
              help='''Task whose ended date is anterior will be archived.
                      Default to current month's first day.''')
@click.option('--after', '-a', default=None,
              help='''Task whose ended date is after the specified date will
                      be archived. Default to prior month's first day.''')
@click.option('--batch-index', default=1000, type=click.INT,
              help='Batch size of tasks to read from db to archive')
@click.option('--bulk-index', default=200, type=click.INT,
              help='Batch size of tasks to bulk index')
@click.option('--batch-clean', default=1000, type=click.INT,
              help='Batch size of task to clean after archival')
@click.option('--dry-run/--no-dry-run', is_flag=True, default=False,
              help='Default to list only what would be archived.')
@click.option('--verbose', is_flag=True, default=False,
              help='Default to list only what would be archived.')
@click.option('--cleanup/--no-cleanup', is_flag=True, default=True,
              help='Clean up archived tasks (default)')
@click.option('--start-from', type=click.INT, default=-1,
              help='(Optional) default task id to start from. Default is -1.')
@click.pass_context
def archive_tasks(ctx, before, after, batch_index, bulk_index, batch_clean,
                  dry_run, verbose, cleanup, start_from):
    """Archive task/task_run whose (task_type is 'oneshot' and task_status
       is 'completed') or (task_type is 'recurring' and task_status is
       'disabled').

       With --dry-run flag set (default), only list those.

    """
    es_client = SWHElasticSearchClient()
    logging.basicConfig(level=logging.DEBUG if verbose else logging.INFO)
    log = logging.getLogger('swh.scheduler.cli.archive')
    logging.getLogger('urllib3').setLevel(logging.WARN)
    logging.getLogger('elasticsearch').setLevel(logging.WARN)
    if dry_run:
        log.info('**DRY-RUN** (only reading db)')
    if not cleanup:
        log.info('**NO CLEANUP**')

    now = arrow.utcnow()

    # Default to archive tasks from a rolling month starting the week
    # prior to the current one
    if not before:
        before = now.shift(weeks=-1).format('YYYY-MM-DD')

    if not after:
        after = now.shift(weeks=-1).shift(months=-1).format('YYYY-MM-DD')

    log.debug('index: %s; cleanup: %s; period: [%s ; %s]' % (
        not dry_run, not dry_run and cleanup, after, before))

    def group_by_index_name(data, es_client=es_client):
        """Given a data record, determine the index's name through its ending
           date. This varies greatly depending on the task_run's
           status.

        """
        date = data.get('started')
        if not date:
            date = data['scheduled']
        return es_client.compute_index_name(date.year, date.month)

    def index_data(before, last_id, batch_index, backend=ctx.obj):
        tasks_in = backend.filter_task_to_archive(
            after, before, last_id=last_id, limit=batch_index)
        for index_name, tasks_group in itertools.groupby(
                tasks_in, key=group_by_index_name):
            log.debug('Index tasks to %s' % index_name)
            if dry_run:
                for task in tasks_group:
                    yield task
                continue

            yield from es_client.streaming_bulk(
                index_name, tasks_group, source=['task_id', 'task_run_id'],
                chunk_size=bulk_index, log=log)

    gen = index_data(before, last_id=start_from, batch_index=batch_index)
    if cleanup:
        for task_ids in utils.grouper(gen, n=batch_clean):
            task_ids = list(task_ids)
            log.info('Clean up %s tasks: [%s, ...]' % (
                len(task_ids), task_ids[0]))
            if dry_run:  # no clean up
                continue
            ctx.obj.delete_archived_tasks(task_ids)
    else:
        for task_ids in utils.grouper(gen, n=batch_index):
            task_ids = list(task_ids)
            log.info('Indexed %s tasks: [%s, ...]' % (
                len(task_ids), task_ids[0]))


@cli.group('task-run')
@click.pass_context
def task_run(ctx):
    """Manipulate task runs."""
    pass


if __name__ == '__main__':
    cli()



import os


def setup_celery():
    os.environ.setdefault('CELERY_BROKER_URL', 'memory://')
    os.environ.setdefault('CELERY_RESULT_BACKEND', 'cache+memory://')


class CeleryTestFixture:
    """Mix this in a test subject class to setup Celery config for testing
    purpose.

    Can be overriden by CELERY_BROKER_URL and CELERY_RESULT_BACKEND env vars.
    """

    def setUp(self):
        setup_celery()
        super().setUp()



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import binascii
import datetime
from functools import wraps
import json
import tempfile

from arrow import Arrow, utcnow
import psycopg2
import psycopg2.extras
from psycopg2.extensions import AsIs

from swh.core.config import SWHConfig


def adapt_arrow(arrow):
    return AsIs("'%s'::timestamptz" % arrow.isoformat())


psycopg2.extensions.register_adapter(dict, psycopg2.extras.Json)
psycopg2.extensions.register_adapter(Arrow, adapt_arrow)


def autocommit(fn):
    @wraps(fn)
    def wrapped(self, *args, **kwargs):
        autocommit = False
        if 'cursor' not in kwargs or not kwargs['cursor']:
            autocommit = True
            kwargs['cursor'] = self.cursor()

        try:
            ret = fn(self, *args, **kwargs)
        except Exception:
            if autocommit:
                self.rollback()
            raise

        if autocommit:
            self.commit()

        return ret

    return wrapped


class DbBackend:
    """Mixin intended to be used within scheduling db backend classes

    cf. swh.scheduler.backend.SchedulerBackend, and
    swh.scheduler.updater.backend.SchedulerUpdaterBackend

    """
    def reconnect(self):
        if not self.db or self.db.closed:
            self.db = psycopg2.connect(
                dsn=self.db_conn_dsn,
                cursor_factory=psycopg2.extras.RealDictCursor,
            )

    def cursor(self):
        """Return a fresh cursor on the database, with auto-reconnection in
        case of failure

        """
        cur = None

        # Get a fresh cursor and reconnect at most three times
        tries = 0
        while True:
            tries += 1
            try:
                cur = self.db.cursor()
                cur.execute('select 1')
                break
            except psycopg2.OperationalError:
                if tries < 3:
                    self.reconnect()
                else:
                    raise

        return cur

    def commit(self):
        """Commit a transaction"""
        self.db.commit()

    def rollback(self):
        """Rollback a transaction"""
        self.db.rollback()

    def close_connection(self):
        """Close db connection"""
        if self.db and not self.db.closed:
            self.db.close()

    def _format_query(self, query, keys):
        """Format a query with the given keys"""

        query_keys = ', '.join(keys)
        placeholders = ', '.join(['%s'] * len(keys))

        return query.format(keys=query_keys, placeholders=placeholders)

    def _format_multiquery(self, query, keys, values):
        """Format a query with placeholders generated for multiple values"""
        query_keys = ', '.join(keys)
        placeholders = '), ('.join(
            [', '.join(['%s'] * len(keys))] * len(values)
        )
        ret_values = sum([[value[key] for key in keys]
                          for value in values], [])

        return (
            query.format(keys=query_keys, placeholders=placeholders),
            ret_values,
        )

    def copy_to(self, items, tblname, columns, default_columns={},
                cursor=None, item_cb=None):
        def escape(data):
            if data is None:
                return ''
            if isinstance(data, bytes):
                return '\\x%s' % binascii.hexlify(data).decode('ascii')
            elif isinstance(data, str):
                return '"%s"' % data.replace('"', '""')
            elif isinstance(data, (datetime.datetime, Arrow)):
                # We escape twice to make sure the string generated by
                # isoformat gets escaped
                return escape(data.isoformat())
            elif isinstance(data, dict):
                return escape(json.dumps(data))
            elif isinstance(data, list):
                return escape("{%s}" % ','.join(escape(d) for d in data))
            elif isinstance(data, psycopg2.extras.Range):
                # We escape twice here too, so that we make sure
                # everything gets passed to copy properly
                return escape(
                    '%s%s,%s%s' % (
                        '[' if data.lower_inc else '(',
                        '-infinity' if data.lower_inf else escape(data.lower),
                        'infinity' if data.upper_inf else escape(data.upper),
                        ']' if data.upper_inc else ')',
                    )
                )
            else:
                # We don't escape here to make sure we pass literals properly
                return str(data)
        with tempfile.TemporaryFile('w+') as f:
            for d in items:
                if item_cb is not None:
                    item_cb(d)
                line = []
                for k in columns:
                    v = d.get(k)
                    if not v:
                        v = default_columns.get(k)
                    v = escape(v)
                    line.append(v)
                f.write(','.join(line))
                f.write('\n')
            f.seek(0)
            cursor.copy_expert('COPY %s (%s) FROM STDIN CSV' % (
                tblname, ', '.join(columns)), f)


class SchedulerBackend(SWHConfig, DbBackend):
    """Backend for the Software Heritage scheduling database.

    """
    CONFIG_BASE_FILENAME = 'scheduler'
    DEFAULT_CONFIG = {
        'scheduling_db': ('str', 'dbname=softwareheritage-scheduler-dev'),
    }

    def __init__(self, **override_config):
        super().__init__()
        self.config = self.parse_config_file(global_config=False)
        self.config.update(override_config)
        self.db = None
        self.db_conn_dsn = self.config['scheduling_db']
        self.reconnect()

    task_type_keys = [
        'type', 'description', 'backend_name', 'default_interval',
        'min_interval', 'max_interval', 'backoff_factor', 'max_queue_length',
        'num_retries', 'retry_delay',
    ]

    @autocommit
    def create_task_type(self, task_type, cursor=None):
        """Create a new task type ready for scheduling.

        Args:
            task_type (dict): a dictionary with the following keys:

                - type (str): an identifier for the task type
                - description (str): a human-readable description of what the
                  task does
                - backend_name (str): the name of the task in the
                  job-scheduling backend
                - default_interval (datetime.timedelta): the default interval
                  between two task runs
                - min_interval (datetime.timedelta): the minimum interval
                  between two task runs
                - max_interval (datetime.timedelta): the maximum interval
                  between two task runs
                - backoff_factor (float): the factor by which the interval
                  changes at each run
                - max_queue_length (int): the maximum length of the task queue
                  for this task type

        """
        query = self._format_query(
            """insert into task_type ({keys}) values ({placeholders})""",
            self.task_type_keys,
        )
        cursor.execute(query, [task_type[key] for key in self.task_type_keys])

    @autocommit
    def get_task_type(self, task_type_name, cursor=None):
        """Retrieve the task type with id task_type_name"""
        query = self._format_query(
            "select {keys} from task_type where type=%s",
            self.task_type_keys,
        )
        cursor.execute(query, (task_type_name,))

        ret = cursor.fetchone()

        return ret

    @autocommit
    def get_task_types(self, cursor=None):
        query = self._format_query(
            "select {keys} from task_type",
            self.task_type_keys,
        )
        cursor.execute(query)
        ret = cursor.fetchall()
        return ret

    task_create_keys = [
        'type', 'arguments', 'next_run', 'policy', 'status', 'retries_left',
        'priority'
    ]
    task_keys = task_create_keys + ['id', 'current_interval', 'status']

    @autocommit
    def create_tasks(self, tasks, policy='recurring', cursor=None):
        """Create new tasks.

        Args:
            tasks (list): each task is a dictionary with the following keys:

                - type (str): the task type
                - arguments (dict): the arguments for the task runner, keys:

                      - args (list of str): arguments
                      - kwargs (dict str -> str): keyword arguments

                - next_run (datetime.datetime): the next scheduled run for the
                  task

        Returns:
            a list of created tasks.

        """
        cursor.execute('select swh_scheduler_mktemp_task()')
        self.copy_to(tasks, 'tmp_task', self.task_create_keys,
                     default_columns={
                         'policy': policy,
                         'status': 'next_run_not_scheduled'
                     },
                     cursor=cursor)
        query = self._format_query(
            'select {keys} from swh_scheduler_create_tasks_from_temp()',
            self.task_keys,
        )
        cursor.execute(query)
        return cursor.fetchall()

    @autocommit
    def set_status_tasks(self, task_ids, status='disabled', cursor=None):
        """Set the tasks' status whose ids are listed."""
        if not task_ids:
            return
        query = "UPDATE task SET status = %s WHERE id IN %s"
        cursor.execute(query, (status, tuple(task_ids),))
        return None

    @autocommit
    def disable_tasks(self, task_ids, cursor=None):
        """Disable the tasks whose ids are listed."""
        return self.set_status_tasks(task_ids)

    @autocommit
    def get_tasks(self, task_ids, cursor=None):
        """Retrieve the info of tasks whose ids are listed."""
        query = self._format_query('select {keys} from task where id in %s',
                                   self.task_keys)
        cursor.execute(query, (tuple(task_ids),))
        return cursor.fetchall()

    @autocommit
    def peek_ready_tasks(self, task_type, timestamp=None, num_tasks=None,
                         num_tasks_priority=None,
                         cursor=None):
        """Fetch the list of ready tasks

        Args:
            task_type (str): filtering task per their type
            timestamp (datetime.datetime): peek tasks that need to be executed
                before that timestamp
            num_tasks (int): only peek at num_tasks tasks (with no priority)
            num_tasks_priority (int): only peek at num_tasks_priority
                                      tasks (with priority)

        Returns:
            a list of tasks

        """
        if timestamp is None:
            timestamp = utcnow()

        cursor.execute(
            '''select * from swh_scheduler_peek_ready_tasks(
                %s, %s, %s :: bigint, %s :: bigint)''',
            (task_type, timestamp, num_tasks, num_tasks_priority)
        )

        return cursor.fetchall()

    @autocommit
    def grab_ready_tasks(self, task_type, timestamp=None, num_tasks=None,
                         num_tasks_priority=None, cursor=None):
        """Fetch the list of ready tasks, and mark them as scheduled

        Args:
            task_type (str): filtering task per their type
            timestamp (datetime.datetime): grab tasks that need to be executed
                before that timestamp
            num_tasks (int): only grab num_tasks tasks (with no priority)
            num_tasks_priority (int): only grab oneshot num_tasks tasks (with
                                      priorities)

        Returns:
            a list of tasks

        """
        if timestamp is None:
            timestamp = utcnow()

        cursor.execute(
            '''select * from swh_scheduler_grab_ready_tasks(
                 %s, %s, %s :: bigint, %s :: bigint)''',
            (task_type, timestamp, num_tasks, num_tasks_priority)
        )

        return cursor.fetchall()

    task_run_create_keys = ['task', 'backend_id', 'scheduled', 'metadata']

    @autocommit
    def schedule_task_run(self, task_id, backend_id, metadata=None,
                          timestamp=None, cursor=None):
        """Mark a given task as scheduled, adding a task_run entry in the database.

        Args:
            task_id (int): the identifier for the task being scheduled
            backend_id (str): the identifier of the job in the backend
            metadata (dict): metadata to add to the task_run entry
            timestamp (datetime.datetime): the instant the event occurred

        Returns:
            a fresh task_run entry

        """

        if metadata is None:
            metadata = {}

        if timestamp is None:
            timestamp = utcnow()

        cursor.execute(
            'select * from swh_scheduler_schedule_task_run(%s, %s, %s, %s)',
            (task_id, backend_id, metadata, timestamp)
        )

        return cursor.fetchone()

    @autocommit
    def mass_schedule_task_runs(self, task_runs, cursor=None):
        """Schedule a bunch of task runs.

        Args:
            task_runs (list): a list of dicts with keys:

                - task (int): the identifier for the task being scheduled
                - backend_id (str): the identifier of the job in the backend
                - metadata (dict): metadata to add to the task_run entry
                - scheduled (datetime.datetime): the instant the event occurred

        Returns:
            None
        """
        cursor.execute('select swh_scheduler_mktemp_task_run()')
        self.copy_to(task_runs, 'tmp_task_run', self.task_run_create_keys,
                     cursor=cursor)
        cursor.execute('select swh_scheduler_schedule_task_run_from_temp()')

    @autocommit
    def start_task_run(self, backend_id, metadata=None, timestamp=None,
                       cursor=None):
        """Mark a given task as started, updating the corresponding task_run
           entry in the database.

        Args:
            backend_id (str): the identifier of the job in the backend
            metadata (dict): metadata to add to the task_run entry
            timestamp (datetime.datetime): the instant the event occurred

        Returns:
            the updated task_run entry

        """

        if metadata is None:
            metadata = {}

        if timestamp is None:
            timestamp = utcnow()

        cursor.execute(
            'select * from swh_scheduler_start_task_run(%s, %s, %s)',
            (backend_id, metadata, timestamp)
        )

        return cursor.fetchone()

    @autocommit
    def end_task_run(self, backend_id, status, metadata=None, timestamp=None,
                     result=None, cursor=None):
        """Mark a given task as ended, updating the corresponding task_run entry in the
        database.

        Args:
            backend_id (str): the identifier of the job in the backend
            status (str): how the task ended; one of: 'eventful', 'uneventful',
                'failed'
            metadata (dict): metadata to add to the task_run entry
            timestamp (datetime.datetime): the instant the event occurred

        Returns:
            the updated task_run entry

        """

        if metadata is None:
            metadata = {}

        if timestamp is None:
            timestamp = utcnow()

        cursor.execute(
            'select * from swh_scheduler_end_task_run(%s, %s, %s, %s)',
            (backend_id, status, metadata, timestamp)
        )

        return cursor.fetchone()

    @autocommit
    def filter_task_to_archive(self, after_ts, before_ts, limit=10, last_id=-1,
                               cursor=None):
        """Returns the list of task/task_run prior to a given date to archive.

        """
        last_task_run_id = None
        while True:
            row = None
            cursor.execute(
                "select * from swh_scheduler_task_to_archive(%s, %s, %s, %s)",
                (after_ts, before_ts, last_id, limit)
            )
            for row in cursor:
                # nested type index does not accept bare values
                # transform it as a dict to comply with this
                row['arguments']['args'] = {
                    i: v for i, v in enumerate(row['arguments']['args'])
                }
                kwargs = row['arguments']['kwargs']
                row['arguments']['kwargs'] = json.dumps(kwargs)
                yield row

            if not row:
                break
            _id = row.get('task_id')
            _task_run_id = row.get('task_run_id')
            if last_id == _id and last_task_run_id == _task_run_id:
                break
            last_id = _id
            last_task_run_id = _task_run_id

    @autocommit
    def delete_archived_tasks(self, task_ids, cursor=None):
        """Delete archived tasks as much as possible. Only the task_ids whose
           complete associated task_run have been cleaned up will be.

        """
        _task_ids = _task_run_ids = []
        for task_id in task_ids:
            _task_ids.append(task_id['task_id'])
            _task_run_ids.append(task_id['task_run_id'])

        cursor.execute(
            "select * from swh_scheduler_delete_archived_tasks(%s, %s)",
            (_task_ids, _task_run_ids))



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import click
import logging
import time

from arrow import utcnow

from swh.core.config import SWHConfig
from swh.core import utils
from swh.scheduler import get_scheduler
from swh.scheduler.utils import create_oneshot_task_dict
from swh.scheduler.updater.backend import SchedulerUpdaterBackend


class UpdaterWriter(SWHConfig):
    """Updater writer in charge of updating the scheduler db with latest
       prioritized oneshot tasks

       In effect, this:
       - reads the events from scheduler updater's db
       - converts those events into priority oneshot tasks
       - dumps them into the scheduler db

    """
    CONFIG_BASE_FILENAME = 'backend/scheduler-updater-writer'
    DEFAULT_CONFIG = {
        # access to the scheduler backend
        'scheduler': ('dict', {
            'cls': 'local',
            'args': {
                'scheduling_db': 'dbname=softwareheritage-scheduler-dev',
            },
        }),
        # access to the scheduler updater cache
        'scheduler_updater': ('dict', {
            'scheduling_updater_db':
            'dbname=softwareheritage-scheduler-updater-dev',
            'cache_read_limit': 1000,
        }),
        # waiting time between db reads
        'pause': ('int', 10),
        # verbose or not
        'verbose': ('bool', False),
    }

    def __init__(self, **config):
        if config:
            self.config = config
        else:
            self.config = self.parse_config_file()

        self.scheduler_updater_backend = SchedulerUpdaterBackend(
            **self.config['scheduler_updater'])
        self.scheduler_backend = get_scheduler(**self.config['scheduler'])
        self.pause = self.config['pause']
        self.log = logging.getLogger(
            'swh.scheduler.updater.writer.UpdaterWriter')
        self.log.setLevel(
            logging.DEBUG if self.config['verbose'] else logging.INFO)

    def convert_to_oneshot_task(self, event):
        """Given an event, convert it into oneshot task with priority

        Args:
            event (dict): The event to convert to task

        """
        if event['origin_type'] == 'git':
            return create_oneshot_task_dict(
                'origin-update-git',
                event['url'],
                priority='normal')
        self.log.warning('Type %s is not supported for now, only git' % (
            event['origin_type'], ))
        return None

    def write_event_to_scheduler(self, events):
        """Write events to the scheduler and yield ids when done"""
        # convert events to oneshot tasks
        oneshot_tasks = filter(lambda e: e is not None,
                               map(self.convert_to_oneshot_task, events))
        # write event to scheduler
        self.scheduler_backend.create_tasks(list(oneshot_tasks))
        for e in events:
            yield e['url']

    def run(self):
        """First retrieve events from cache (including origin_type, cnt),
           then convert them into oneshot tasks with priority, then
           write them to the scheduler db, at last remove them from
           cache.

        """
        while True:
            timestamp = utcnow()
            events = list(self.scheduler_updater_backend.cache_read(timestamp))
            if not events:
                break
            for urls in utils.grouper(self.write_event_to_scheduler(events),
                                      n=100):
                self.scheduler_updater_backend.cache_remove(urls)
            time.sleep(self.pause)


@click.command()
@click.option('--verbose/--no-verbose', '-v', default=False,
              help='Verbose mode')
def main(verbose):
    log = logging.getLogger('swh.scheduler.updater.writer')
    log.addHandler(logging.StreamHandler())
    _loglevel = logging.DEBUG if verbose else logging.INFO
    log.setLevel(_loglevel)

    UpdaterWriter().run()


if __name__ == '__main__':
    main()



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import unittest

from arrow import utcnow
from hypothesis import given
from hypothesis.strategies import sets
import pytest

from swh.core.tests.db_testing import SingleDbTestFixture
from swh.scheduler.tests import SQL_DIR
from swh.scheduler.updater.backend import SchedulerUpdaterBackend
from swh.scheduler.updater.events import SWHEvent

from . import from_regex


@pytest.mark.db
class SchedulerUpdaterBackendTest(SingleDbTestFixture, unittest.TestCase):
    TEST_DB_NAME = 'softwareheritage-scheduler-updater-test'
    TEST_DB_DUMP = os.path.join(SQL_DIR, 'updater', '*.sql')

    def setUp(self):
        super().setUp()
        config = {
            'scheduling_updater_db': 'dbname=' + self.TEST_DB_NAME,
            'cache_read_limit': 1000,
        }
        self.backend = SchedulerUpdaterBackend(**config)

    def _empty_tables(self):
        self.cursor.execute(
            """SELECT table_name FROM information_schema.tables
               WHERE table_schema = %s""", ('public', ))
        tables = set(table for (table,) in self.cursor.fetchall())
        for table in tables:
            self.cursor.execute('truncate table %s cascade' % table)
        self.conn.commit()

    def tearDown(self):
        self.backend.close_connection()
        self._empty_tables()
        super().tearDown()

    @given(sets(
        from_regex(
            r'^https://somewhere[.]org/[a-z0-9]{5,7}/[a-z0-9]{3,10}$'),
        min_size=10, max_size=15))
    def test_cache_read(self, urls):
        def gen_events(urls):
            for url in urls:
                yield SWHEvent({
                    'url': url,
                    'type': 'create',
                    'origin_type': 'git',
                })

        self.backend.cache_put(gen_events(urls))
        r = self.backend.cache_read(timestamp=utcnow())

        self.assertNotEqual(r, [])



# Copyright (C) 2015-2018 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import arrow

from swh.scheduler import get_scheduler, compute_nb_tasks_from
from .config import app as main_app


# Max batch size for tasks
MAX_NUM_TASKS = 10000


def run_ready_tasks(backend, app):
    """Run tasks that are ready

    Args:
        backend (Scheduler): backend to read tasks to schedule
        app (App): Celery application to send tasks to

    Returns:
        A list of dictionaries::

          {
            'task': the scheduler's task id,
            'backend_id': Celery's task id,
            'scheduler': arrow.utcnow()
          }

        The result can be used to block-wait for the tasks' results::

          backend_tasks = run_ready_tasks(self.scheduler, app)
          for task in backend_tasks:
              AsyncResult(id=task['backend_id']).get()

    """
    all_backend_tasks = []
    while True:
        cursor = backend.cursor()
        task_types = {}
        pending_tasks = []
        for task_type in backend.get_task_types(cursor=cursor):
            task_type_name = task_type['type']
            task_types[task_type_name] = task_type
            max_queue_length = task_type['max_queue_length']
            backend_name = task_type['backend_name']
            if max_queue_length:
                try:
                    queue_length = app.get_queue_length(backend_name)
                except ValueError:
                    queue_length = None

                if queue_length is None:
                    # Running without RabbitMQ (probably a test env).
                    num_tasks = MAX_NUM_TASKS
                else:
                    num_tasks = min(max_queue_length - queue_length,
                                    MAX_NUM_TASKS)
            else:
                num_tasks = MAX_NUM_TASKS
            if num_tasks > 0:
                num_tasks, num_tasks_priority = compute_nb_tasks_from(
                    num_tasks)

                pending_tasks.extend(
                    backend.grab_ready_tasks(
                        task_type_name,
                        num_tasks=num_tasks,
                        num_tasks_priority=num_tasks_priority,
                        cursor=cursor))

        if not pending_tasks:
            return all_backend_tasks

        backend_tasks = []
        for task in pending_tasks:
            args = task['arguments']['args']
            kwargs = task['arguments']['kwargs']

            backend_name = task_types[task['type']]['backend_name']
            celery_result = app.send_task(
                backend_name, args=args, kwargs=kwargs,
            )

            data = {
                'task': task['id'],
                'backend_id': celery_result.id,
                'scheduled': arrow.utcnow(),
            }

            backend_tasks.append(data)

        backend.mass_schedule_task_runs(backend_tasks, cursor=cursor)

        backend.commit()

        all_backend_tasks.extend(backend_tasks)


if __name__ == '__main__':
    for module in main_app.conf.CELERY_IMPORTS:
        __import__(module)

    main_backend = get_scheduler('local')
    try:
        run_ready_tasks(main_backend, main_app)
    except Exception:
        main_backend.rollback()
        raise



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import click

from flask import request

from swh.core import config
from swh.scheduler import get_scheduler as get_scheduler_from
from swh.core.api import (SWHServerAPIApp, decode_request,
                          error_handler,
                          encode_data_server as encode_data)

DEFAULT_CONFIG_PATH = 'backend/scheduler'
DEFAULT_CONFIG = {
    'scheduler': ('dict', {
        'cls': 'local',
        'args': {
            'scheduling_db': 'dbname=softwareheritage-scheduler-dev',
        },
    })
}


app = SWHServerAPIApp(__name__)
scheduler = None


@app.errorhandler(Exception)
def my_error_handler(exception):
    return error_handler(exception, encode_data)


def get_sched():
    global scheduler
    if not scheduler:
        scheduler = get_scheduler_from(**app.config['scheduler'])
    return scheduler


@app.route('/')
def index():
    return 'SWH Scheduler API server'


@app.route('/close_connection', methods=['POST'])
def close_connection():
    return encode_data(get_sched().close_connection())


@app.route('/set_status_tasks', methods=['POST'])
def set_status_tasks():
    return encode_data(get_sched().set_status_tasks(**decode_request(request)))


@app.route('/create_task_type', methods=['POST'])
def create_task_type():
    return encode_data(get_sched().create_task_type(**decode_request(request)))


@app.route('/get_task_type', methods=['POST'])
def get_task_type():
    return encode_data(get_sched().get_task_type(**decode_request(request)))


@app.route('/get_task_types', methods=['POST'])
def get_task_types():
    return encode_data(get_sched().get_task_types(**decode_request(request)))


@app.route('/create_tasks', methods=['POST'])
def create_tasks():
    return encode_data(get_sched().create_tasks(**decode_request(request)))


@app.route('/disable_tasks', methods=['POST'])
def disable_tasks():
    return encode_data(get_sched().disable_tasks(**decode_request(request)))


@app.route('/get_tasks', methods=['POST'])
def get_tasks():
    return encode_data(get_sched().get_tasks(**decode_request(request)))


@app.route('/peek_ready_tasks', methods=['POST'])
def peek_ready_tasks():
    return encode_data(get_sched().peek_ready_tasks(**decode_request(request)))


@app.route('/grab_ready_tasks', methods=['POST'])
def grab_ready_tasks():
    return encode_data(get_sched().grab_ready_tasks(**decode_request(request)))


@app.route('/schedule_task_run', methods=['POST'])
def schedule_task_run():
    return encode_data(get_sched().schedule_task_run(
        **decode_request(request)))


@app.route('/mass_schedule_task_runs', methods=['POST'])
def mass_schedule_task_runs():
    return encode_data(
        get_sched().mass_schedule_task_runs(**decode_request(request)))


@app.route('/start_task_run', methods=['POST'])
def start_task_run():
    return encode_data(get_sched().start_task_run(**decode_request(request)))


@app.route('/end_task_run', methods=['POST'])
def end_task_run():
    return encode_data(get_sched().end_task_run(**decode_request(request)))


@app.route('/filter_task_to_archive', methods=['POST'])
def filter_task_to_archive():
    return encode_data(
        get_sched().filter_task_to_archive(**decode_request(request)))


@app.route('/delete_archived_tasks', methods=['POST'])
def delete_archived_tasks():
    return encode_data(
        get_sched().delete_archived_tasks(**decode_request(request)))


def run_from_webserver(environ, start_response,
                       config_path=DEFAULT_CONFIG_PATH):
    """Run the WSGI app from the webserver, loading the configuration."""
    cfg = config.load_named_config(config_path, DEFAULT_CONFIG)
    app.config.update(cfg)
    handler = logging.StreamHandler()
    app.logger.addHandler(handler)
    return app(environ, start_response)


@click.command()
@click.argument('config-path', required=1)
@click.option('--host', default='0.0.0.0',
              help="Host to run the scheduler server api")
@click.option('--port', default=5008, type=click.INT,
              help="Binding port of the server")
@click.option('--debug/--nodebug', default=True,
              help="Indicates if the server should run in debug mode")
def launch(config_path, host, port, debug):
    app.config.update(config.read(config_path, DEFAULT_CONFIG))
    app.run(host, port=port, debug=bool(debug))


if __name__ == '__main__':
    launch()



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from arrow import utcnow
from swh.core.config import SWHConfig
from swh.scheduler.backend import DbBackend, autocommit


class SchedulerUpdaterBackend(SWHConfig, DbBackend):
    CONFIG_BASE_FILENAME = 'backend/scheduler-updater'
    DEFAULT_CONFIG = {
        'scheduling_updater_db': (
            'str', 'dbname=softwareheritage-scheduler-updater-dev'),
        'cache_read_limit': ('int', 1000),
    }

    def __init__(self, **override_config):
        super().__init__()
        if override_config:
            self.config = override_config
        else:
            self.config = self.parse_config_file(global_config=False)
        self.db = None
        self.db_conn_dsn = self.config['scheduling_updater_db']
        self.limit = self.config['cache_read_limit']
        self.reconnect()

    cache_put_keys = ['url', 'cnt', 'last_seen', 'origin_type']

    @autocommit
    def cache_put(self, events, timestamp=None, cursor=None):
        """Write new events in the backend.

        """
        if timestamp is None:
            timestamp = utcnow()

        def prepare_events(events):
            for e in events:
                event = e.get()
                seen = event['last_seen']
                if seen is None:
                    event['last_seen'] = timestamp
                yield event

        cursor.execute('select swh_mktemp_cache()')
        self.copy_to(prepare_events(events),
                     'tmp_cache', self.cache_put_keys, cursor=cursor)
        cursor.execute('select swh_cache_put()')

    cache_read_keys = ['id', 'url', 'origin_type', 'cnt', 'first_seen',
                       'last_seen']

    @autocommit
    def cache_read(self, timestamp=None, limit=None, cursor=None):
        """Read events from the cache prior to timestamp.

        """
        if not timestamp:
            timestamp = utcnow()

        if not limit:
            limit = self.limit

        q = self._format_query('select {keys} from swh_cache_read(%s, %s)',
                               self.cache_read_keys)
        cursor.execute(q, (timestamp, limit))
        for r in cursor.fetchall():
            r['id'] = r['id'].tobytes()
            yield r

    @autocommit
    def cache_remove(self, entries, cursor=None):
        """Clean events from the cache

        """
        q = 'delete from cache where url in (%s)' % (
            ', '.join(("'%s'" % e for e in entries)), )
        cursor.execute(q)



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import click
import datetime
import logging
import socket

from arrow import utcnow
from kombu import Queue
from celery.events import EventReceiver

from swh.scheduler import get_scheduler
from .config import app as main_app


class ReliableEventReceiver(EventReceiver):
    def __init__(self, channel, handlers=None, routing_key='#',
                 node_id=None, app=None, queue_prefix='celeryev',
                 accept=None):
        super(ReliableEventReceiver, self).__init__(
            channel, handlers, routing_key, node_id, app, queue_prefix, accept)

        self.queue = Queue('.'.join([self.queue_prefix, self.node_id]),
                           exchange=self.exchange,
                           routing_key=self.routing_key,
                           auto_delete=False,
                           durable=True)

    def get_consumers(self, consumer, channel):
        return [consumer(queues=[self.queue],
                         callbacks=[self._receive], no_ack=False,
                         accept=self.accept)]

    def _receive(self, bodies, message):
        logging.debug('## event-receiver: bodies: %s' % bodies)
        logging.debug('## event-receiver: message: %s' % message)
        if not isinstance(bodies, list):  # celery<4 returned body as element
            bodies = [bodies]
        for body in bodies:
            type, body = self.event_from_message(body)
            self.process(type, body, message)

    def process(self, type, event, message):
        """Process the received event by dispatching it to the appropriate
        handler."""
        handler = self.handlers.get(type) or self.handlers.get('*')
        logging.debug('## event-receiver: type: %s' % type)
        logging.debug('## event-receiver: event: %s' % event)
        logging.debug('## event-receiver: handler: %s' % handler)
        handler and handler(event, message)


ACTION_SEND_DELAY = datetime.timedelta(seconds=1.0)
ACTION_QUEUE_MAX_LENGTH = 1000


def event_monitor(app, backend):
    actions = {
        'last_send': utcnow() - 2*ACTION_SEND_DELAY,
        'queue': [],
    }

    def try_perform_actions(actions=actions):
        if not actions['queue']:
            return
        if utcnow() - actions['last_send'] > ACTION_SEND_DELAY or \
           len(actions['queue']) > ACTION_QUEUE_MAX_LENGTH:
            perform_actions(actions)

    def perform_actions(actions, backend=backend):
        action_map = {
            'start_task_run': backend.start_task_run,
            'end_task_run': backend.end_task_run,
        }

        messages = []
        cursor = backend.cursor()
        for action in actions['queue']:
            messages.append(action['message'])
            function = action_map[action['action']]
            args = action.get('args', ())
            kwargs = action.get('kwargs', {})
            kwargs['cursor'] = cursor
            function(*args, **kwargs)

        backend.commit()
        for message in messages:
            message.ack()
        actions['queue'] = []
        actions['last_send'] = utcnow()

    def queue_action(action, actions=actions):
        actions['queue'].append(action)
        try_perform_actions()

    def catchall_event(event, message):
        message.ack()
        try_perform_actions()

    def task_started(event, message):
        logging.debug('#### task_started: event: %s' % event)
        logging.debug('#### task_started: message: %s' % message)

        queue_action({
            'action': 'start_task_run',
            'args': [event['uuid']],
            'kwargs': {
                'timestamp': utcnow(),
                'metadata': {
                    'worker': event['hostname'],
                },
            },
            'message': message,
        })

    def task_succeeded(event, message):
        logging.debug('#### task_succeeded: event: %s' % event)
        logging.debug('#### task_succeeded: message: %s' % message)
        result = event['result']

        logging.debug('#### task_succeeded: result: %s' % result)
        try:
            status = result.get('status')
            if status == 'success':
                status = 'eventful' if result.get('eventful') else 'uneventful'
        except Exception:
            status = 'eventful' if result else 'uneventful'

        queue_action({
            'action': 'end_task_run',
            'args': [event['uuid']],
            'kwargs': {
                'timestamp': utcnow(),
                'status': status,
                'result': result,
            },
            'message': message,
        })

    def task_failed(event, message):
        logging.debug('#### task_failed: event: %s' % event)
        logging.debug('#### task_failed: message: %s' % message)

        queue_action({
            'action': 'end_task_run',
            'args': [event['uuid']],
            'kwargs': {
                'timestamp': utcnow(),
                'status': 'failed',
            },
            'message': message,
        })

    recv = ReliableEventReceiver(
        main_app.connection(),
        app=main_app,
        handlers={
            'task-started': task_started,
            'task-result': task_succeeded,
            'task-failed': task_failed,
            '*': catchall_event,
        },
        node_id='listener-%s' % socket.gethostname(),
    )

    recv.capture(limit=None, timeout=None, wakeup=True)


@click.command()
@click.option('--cls', '-c', default='local',
              help="Scheduler's class, default to 'local'")
@click.option(
    '--database', '-d', help='Scheduling database DSN')
@click.option('--url', '-u',
              help="(Optional) Scheduler's url access")
@click.option('--verbose', is_flag=True, default=False,
              help='Default to be silent')
def main(cls, database, url, verbose):
    if verbose:
        logging.basicConfig(level=logging.DEBUG if verbose else logging.INFO)

    scheduler = None
    override_config = {}
    if cls == 'local':
        if database:
            override_config = {'scheduling_db': database}
        scheduler = get_scheduler(cls, args=override_config)
    elif cls == 'remote':
        if url:
            override_config = {'url': url}
        scheduler = get_scheduler(cls, args=override_config)

    if not scheduler:
        raise ValueError('Scheduler class (local/remote) must be instantiated')

    event_monitor(main_app, backend=scheduler)


if __name__ == '__main__':
    main()



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import itertools
import importlib
import logging
import os
import urllib.parse

from celery import Celery
from celery.signals import setup_logging, celeryd_after_setup
from celery.utils.log import ColorFormatter
from celery.worker.control import Panel

from kombu import Exchange, Queue
from kombu.five import monotonic as _monotonic

import requests

from swh.scheduler.task import Task

from swh.core.config import load_named_config
from swh.core.logger import JournalHandler

DEFAULT_CONFIG_NAME = 'worker'
CONFIG_NAME_ENVVAR = 'SWH_WORKER_INSTANCE'
CONFIG_NAME_TEMPLATE = 'worker/%s'

DEFAULT_CONFIG = {
    'task_broker': ('str', 'amqp://guest@localhost//'),
    'task_modules': ('list[str]', []),
    'task_queues': ('list[str]', []),
    'task_soft_time_limit': ('int', 0),
}


@setup_logging.connect
def setup_log_handler(loglevel=None, logfile=None, format=None,
                      colorize=None, **kwargs):
    """Setup logging according to Software Heritage preferences.

    We use the command-line loglevel for tasks only, as we never
    really care about the debug messages from celery.
    """

    if loglevel is None:
        loglevel = logging.DEBUG

    formatter = logging.Formatter(format)

    root_logger = logging.getLogger('')
    root_logger.setLevel(logging.INFO)

    if loglevel == logging.DEBUG:
        color_formatter = ColorFormatter(format) if colorize else formatter
        console = logging.StreamHandler()
        console.setLevel(logging.DEBUG)
        console.setFormatter(color_formatter)
        root_logger.addHandler(console)

    systemd_journal = JournalHandler()
    systemd_journal.setLevel(logging.DEBUG)
    systemd_journal.setFormatter(formatter)
    root_logger.addHandler(systemd_journal)

    celery_logger = logging.getLogger('celery')
    celery_logger.setLevel(logging.INFO)

    # Silence useless "Starting new HTTP connection" messages
    urllib3_logger = logging.getLogger('urllib3')
    urllib3_logger.setLevel(logging.WARNING)

    swh_logger = logging.getLogger('swh')
    swh_logger.setLevel(loglevel)

    # get_task_logger makes the swh tasks loggers children of celery.task
    celery_task_logger = logging.getLogger('celery.task')
    celery_task_logger.setLevel(loglevel)


@celeryd_after_setup.connect
def setup_queues_and_tasks(sender, instance, **kwargs):
    """Signal called on worker start.

    This automatically registers swh.scheduler.task.Task subclasses as
    available celery tasks.

    This also subscribes the worker to the "implicit" per-task queues defined
    for these task classes.

    """

    for module_name in itertools.chain(
            # celery worker -I flag
            instance.app.conf['include'],
            # set from the celery / swh worker instance configuration file
            instance.app.conf['imports'],
    ):
        module = importlib.import_module(module_name)
        for name in dir(module):
            obj = getattr(module, name)
            if (
                    isinstance(obj, type)
                    and issubclass(obj, Task)
                    and obj != Task  # Don't register the abstract class itself
            ):
                class_name = '%s.%s' % (module_name, name)
                instance.app.register_task_class(class_name, obj)

    for task_name in instance.app.tasks:
        if task_name.startswith('swh.'):
            instance.app.amqp.queues.select_add(task_name)


@Panel.register
def monotonic(state):
    """Get the current value for the monotonic clock"""
    return {'monotonic': _monotonic()}


class TaskRouter:
    """Route tasks according to the task_queue attribute in the task class"""
    def route_for_task(self, task, *args, **kwargs):
        if task.startswith('swh.'):
            return {'queue': task}


class CustomCelery(Celery):
    def get_queue_stats(self, queue_name):
        """Get the statistics regarding a queue on the broker.

        Arguments:
          queue_name: name of the queue to check

        Returns a dictionary raw from the RabbitMQ management API;
        or `None` if the current configuration does not use RabbitMQ.

        Interesting keys:
         - consumers (number of consumers for the queue)
         - messages (number of messages in queue)
         - messages_unacknowledged (number of messages currently being
           processed)

        Documentation: https://www.rabbitmq.com/management.html#http-api
        """

        conn_info = self.connection().info()
        if conn_info['transport'] == 'memory':
            # We're running in a test environment, without RabbitMQ.
            return None
        url = 'http://{hostname}:{port}/api/queues/{vhost}/{queue}'.format(
            hostname=conn_info['hostname'],
            port=conn_info['port'] + 10000,
            vhost=urllib.parse.quote(conn_info['virtual_host'], safe=''),
            queue=urllib.parse.quote(queue_name, safe=''),
        )
        credentials = (conn_info['userid'], conn_info['password'])
        r = requests.get(url, auth=credentials)
        if r.status_code == 404:
            return {}
        if r.status_code != 200:
            raise ValueError('Got error %s when reading queue stats: %s' % (
                r.status_code, r.json()))
        return r.json()

    def get_queue_length(self, queue_name):
        """Shortcut to get a queue's length"""
        stats = self.get_queue_stats(queue_name)
        if stats:
            return stats.get('messages')

    def register_task_class(self, name, cls):
        """Register a class-based task under the given name"""
        if name in self.tasks:
            return

        task_instance = cls()
        task_instance.name = name
        self.register_task(task_instance)


INSTANCE_NAME = os.environ.get(CONFIG_NAME_ENVVAR)
if INSTANCE_NAME:
    CONFIG_NAME = CONFIG_NAME_TEMPLATE % INSTANCE_NAME
else:
    CONFIG_NAME = DEFAULT_CONFIG_NAME

# Load the Celery config
CONFIG = load_named_config(CONFIG_NAME, DEFAULT_CONFIG)

# Celery Queues
CELERY_QUEUES = [Queue('celery', Exchange('celery'), routing_key='celery')]

for queue in CONFIG['task_queues']:
    CELERY_QUEUES.append(Queue(queue, Exchange(queue), routing_key=queue))

# Instantiate the Celery app
app = CustomCelery()
app.conf.update(
    # The broker
    broker_url=CONFIG['task_broker'],
    # Timezone configuration: all in UTC
    enable_utc=True,
    timezone='UTC',
    # Imported modules
    imports=CONFIG['task_modules'],
    # Time (in seconds, or a timedelta object) for when after stored task
    # tombstones will be deleted. None means to never expire results.
    result_expires=None,
    # A string identifying the default serialization method to use. Can
    # be json (default), pickle, yaml, msgpack, or any custom
    # serialization methods that have been registered with
    task_serializer='msgpack',
    # Result serialization format
    result_serializer='msgpack',
    # Late ack means the task messages will be acknowledged after the task has
    # been executed, not just before, which is the default behavior.
    task_acks_late=True,
    # A string identifying the default serialization method to use.
    # Can be pickle (default), json, yaml, msgpack or any custom serialization
    # methods that have been registered with kombu.serialization.registry
    accept_content=['msgpack', 'json'],
    # If True the task will report its status as “started”
    # when the task is executed by a worker.
    task_track_started=True,
    # Default compression used for task messages. Can be gzip, bzip2
    # (if available), or any custom compression schemes registered
    # in the Kombu compression registry.
    # result_compression='bzip2',
    # task_compression='bzip2',
    # Disable all rate limits, even if tasks has explicit rate limits set.
    # (Disabling rate limits altogether is recommended if you don’t have any
    # tasks using them.)
    worker_disable_rate_limits=True,
    # Task hard time limit in seconds. The worker processing the task will be
    # killed and replaced with a new one when this is exceeded.
    # task_time_limit=3600,
    # Task soft time limit in seconds.
    # The SoftTimeLimitExceeded exception will be raised when this is exceeded.
    # The task can catch this to e.g. clean up before the hard time limit
    # comes.
    task_soft_time_limit=CONFIG['task_soft_time_limit'],
    # Task routing
    task_routes=TaskRouter(),
    # Task queues this worker will consume from
    task_queues=CELERY_QUEUES,
    # Allow pool restarts from remote
    worker_pool_restarts=True,
    # Do not prefetch tasks
    worker_prefetch_multiplier=1,
    # Send events
    worker_send_task_events=True,
    # Do not send useless task_sent events
    task_send_sent_event=False,
)



# Copyright (C) 2015-2017 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import celery.app.task
from celery.utils.log import get_task_logger


class Task(celery.app.task.Task):
    """a schedulable task (abstract class)

    Sub-classes must implement the run_task() method.

    Current implementation is based on Celery. See
    http://docs.celeryproject.org/en/latest/reference/celery.app.task.html for
    how to use tasks once instantiated

    """

    abstract = True

    def run(self, *args, **kwargs):
        """This method is called by the celery worker when a task is received.

        Should not be overridden as we need our special events to be sent for
        the reccurrent scheduler. Override run_task instead."""
        try:
            result = self.run_task(*args, **kwargs)
        except Exception as e:
            self.send_event('task-result-exception')
            raise e from None
        else:
            self.send_event('task-result', result=result)
            return result

    def run_task(self, *args, **kwargs):
        """Perform the task.

        Must return a json-serializable value as it is passed back to the task
        scheduler using a celery event.
        """
        raise NotImplementedError('tasks must implement the run_task() method')

    @property
    def log(self):
        if not hasattr(self, '__log'):
            self.__log = get_task_logger('%s.%s' %
                                         (__name__, self.__class__.__name__))
        return self.__log



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from swh.scheduler.tests.scheduler_testing import SchedulerTestFixture
from swh.scheduler.task import Task
from swh.scheduler.utils import create_task_dict

task_has_run = False


class SomeTestTask(Task):
    def run(self, *, foo):
        global task_has_run
        assert foo == 'bar'
        task_has_run = True


class FixtureTest(SchedulerTestFixture, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.add_scheduler_task_type(
            'some_test_task_type',
            'swh.scheduler.tests.test_fixtures.SomeTestTask',
            SomeTestTask,
        )

    def test_task_run(self):
        self.scheduler.create_tasks([create_task_dict(
            'some_test_task_type',
            'oneshot',
            foo='bar',
            )])
        self.assertEqual(task_has_run, False)
        self.run_ready_tasks()
        self.assertEqual(task_has_run, True)



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import click
import logging

from swh.scheduler.updater.ghtorrent import GHTorrentConsumer


@click.command()
@click.option('--verbose/--no-verbose', '-v', default=False,
              help='Verbose mode')
def main(verbose):
    """Consume events from ghtorrent and write them to cache.

    """
    log = logging.getLogger('swh.scheduler.updater.ghtorrent.cli')
    log.addHandler(logging.StreamHandler())
    _loglevel = logging.DEBUG if verbose else logging.INFO
    log.setLevel(_loglevel)

    GHTorrentConsumer().run()


if __name__ == '__main__':
    main()



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from celery import current_app as app

from swh.scheduler import task
from .celery_testing import CeleryTestFixture


class Task(CeleryTestFixture, unittest.TestCase):

    def test_not_implemented_task(self):
        class NotImplementedTask(task.Task):
            name = 'NotImplementedTask'

            pass

        app.register_task(NotImplementedTask())

        with self.assertRaises(NotImplementedError):
            NotImplementedTask().run()

    def test_add_task(self):
        class AddTask(task.Task):
            name = 'AddTask'

            def run_task(self, x, y):
                return x + y

        app.register_task(AddTask())

        r = AddTask().apply([2, 3])
        self.assertTrue(r.successful())
        self.assertEqual(r.result, 5)



import glob
import os.path
import datetime

from celery.result import AsyncResult
from celery.contrib.testing.worker import start_worker
import celery.contrib.testing.tasks  # noqa           
import pytest

from swh.core.tests.db_testing import DbTestFixture, DB_DUMP_TYPES
from swh.core.utils import numfile_sortkey as sortkey

from swh.scheduler import get_scheduler
from swh.scheduler.celery_backend.runner import run_ready_tasks
from swh.scheduler.celery_backend.config import app
from swh.scheduler.tests.celery_testing import CeleryTestFixture

from . import SQL_DIR

DUMP_FILES = os.path.join(SQL_DIR, '*.sql')


@pytest.mark.db
class SchedulerTestFixture(CeleryTestFixture, DbTestFixture):
    """Base class for test case classes, providing an SWH scheduler as
    the `scheduler` attribute."""
    SCHEDULER_DB_NAME = 'softwareheritage-scheduler-test-fixture'

    def add_scheduler_task_type(self, task_type, backend_name,
                                task_class=None):
        task_type = {
            'type': task_type,
            'description': 'Update a git repository',
            'backend_name': backend_name,
            'default_interval': datetime.timedelta(days=64),
            'min_interval': datetime.timedelta(hours=12),
            'max_interval': datetime.timedelta(days=64),
            'backoff_factor': 2,
            'max_queue_length': None,
            'num_retries': 7,
            'retry_delay': datetime.timedelta(hours=2),
        }
        self.scheduler.create_task_type(task_type)
        if task_class:
            app.register_task_class(backend_name, task_class)

    def run_ready_tasks(self):
        """Runs the scheduler and a Celery worker, then blocks until
        all tasks are completed."""

        # Make sure the worker is listening to all task-specific queues
        for task in self.scheduler.get_task_types():
            app.amqp.queues.select_add(task['backend_name'])

        with start_worker(app):
            backend_tasks = run_ready_tasks(self.scheduler, app)
            for task in backend_tasks:
                # Make sure the task completed
                AsyncResult(id=task['backend_id']).get()

    @classmethod
    def setUpClass(cls):
        all_dump_files = sorted(glob.glob(DUMP_FILES), key=sortkey)

        all_dump_files = [(x, DB_DUMP_TYPES[os.path.splitext(x)[1]])
                          for x in all_dump_files]

        cls.add_db(name=cls.SCHEDULER_DB_NAME,
                   dumps=all_dump_files)
        super().setUpClass()

    def setUp(self):
        super().setUp()
        self.scheduler_config = {
                'scheduling_db': 'dbname=' + self.SCHEDULER_DB_NAME}
        self.scheduler = get_scheduler('local', self.scheduler_config)

    def tearDown(self):
        self.scheduler.close_connection()
        super().tearDown()
